"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.Form = void 0;
const utils = require("../utils");
const components = require("../components");
const field_1 = require("./field");
class Form {
    constructor(name) {
        this.getKeys = () => {
            return this.fields.keys();
        };
        this.valueExists = (path) => {
            return utils.exists(this.values, path);
        };
        this.getValue = (path) => {
            return utils.get(this.values, path);
        };
        this.setValue = (path, value) => {
            if (value !== undefined && value !== "")
                utils.set(this.values, path, value);
            else
                utils.remove(this.values, path);
            this.trigger();
        };
        this.setSubmissionErrors = (errors) => {
            const normalizedErrors = utils.lowerCaseKeys(utils.flattenObject(errors));
            const keys = this.getKeys();
            for (const key of keys) {
                if (key in normalizedErrors) {
                    const fieldErrors = typeof normalizedErrors[key] == "string" ? [normalizedErrors[key]] : normalizedErrors[key];
                    this.getField(key.toLowerCase()).setSubmissionErrors(fieldErrors);
                }
                else {
                    const field = this.fields.get(key);
                    field.submitFailed = field.validate() === false;
                    field.trigger();
                }
            }
        };
        this.getField = (name) => {
            const nameLowerCase = name.toLowerCase();
            let field = this.fields.get(nameLowerCase);
            if (field === undefined) {
                field = new field_1.Field(this, nameLowerCase);
                this.fields.set(nameLowerCase, field);
            }
            return field;
        };
        this.removeField = (name) => {
            const nameLowerCase = name.toLowerCase();
            return this.fields.delete(nameLowerCase);
        };
        this.getValues = () => {
            return Object.assign({}, this.values);
        };
        this.reset = () => {
            const keys = this.getKeys();
            for (const key of keys)
                this.fields.get(key).reset();
        };
        this.checkCanSubmit = () => {
            let allValid = true;
            const keys = this.getKeys();
            for (const key of keys) {
                const field = this.fields.get(key);
                field.submitFailed = field.validate() === false;
                if (field.submitFailed) {
                    field.trigger();
                    allValid = false;
                }
            }
            return allValid;
        };
        this.submit = (callback) => {
            if (this.checkCanSubmit()) {
                if (callback)
                    callback(this.values);
                else if (this.onSubmit)
                    this.onSubmit(this.values);
                else
                    console.warn(`Unhandled submit of ${this.name}`, this.values);
            }
        };
        this.subscribe = (subscriber) => {
            this.subscribers.add(subscriber);
        };
        this.unsubscribe = (subscriber) => {
            this.subscribers.delete(subscriber);
        };
        this.trigger = () => {
            for (const subscriber of this.subscribers)
                subscriber(this.getValues());
        };
        this.name = name !== null && name !== void 0 ? name : utils.uuid();
        this.Form = components.getFormComponent(this);
        this.Submit = components.getSubmitComponent(this);
        this.Reset = components.getResetComponent(this);
        this.Field = components.getFieldComponent(this);
        this.Input = components.getInputComponent(this);
        this.FieldArray = components.getFieldArrayComponent(this);
        this.FieldReader = components.getFieldReaderComponent(this);
        this.onSubmit = undefined;
        this.values = {};
        this.fields = new Map();
        this.subscribers = new Set();
    }
}
exports.Form = Form;
//# sourceMappingURL=form.js.map